#include "../std.h"

#include <globus_gsi_credential.h>

#include <vector>

#include <voms_api.h>
#include <Client.h>
#include <api_util.h>
#include "../misc/escaped.h"
#include "../misc/stringtoint.h"
#include "../misc/globus_error_utils.h"
#include "../misc/log_time.h"

#include "auth.h"

const char* voms_error(verror_type err);
int renew_voms_cert(const string &contact,const string &voms_dir,const string &uri);
int process_vomsproxy(const char* filename,vector<struct voms> &data,bool auto_cert = false);

int AuthUser::match_voms(const char* line) {
  // parse line
  string vo("");
  string group("");
  string role("");
  string capabilities("");
  string auto_c("");
  bool auto_cert = false;
  int n;
  n=input_escaped_string(line,vo,' ','"');
  if(n == 0) {
    olog<<"Missing VO in configuration"<<endl; 
    return AAA_FAILURE;
  };
  line+=n;
  n=input_escaped_string(line,group,' ','"');
  if(n == 0) {
    olog<<"Missing group in configuration"<<endl; 
    return AAA_FAILURE;
  };
  line+=n;
  n=input_escaped_string(line,role,' ','"');
  if(n == 0) {
    olog<<"Missing role in configuration"<<endl; 
    return AAA_FAILURE;
  };
  line+=n;
  n=input_escaped_string(line,capabilities,' ','"');
  if(n == 0) {
    olog<<"Missing capabilities in configuration"<<endl; 
    return AAA_FAILURE;
  };
  n=input_escaped_string(line,auto_c,' ','"');
  if(n != 0) {
    if(auto_c == "auto") auto_cert=true;
  };
  // cerr<<"vo: "<<vo<<endl;
  // cerr<<"group: "<<group<<endl;
  // cerr<<"role: "<<role<<endl;
  // cerr<<"capabilities: "<<capabilities<<endl;
  // extract info from voms proxy
  if(voms_data.size() == 0) {
    if(filename.length() > 0) {
      int err = process_vomsproxy(filename.c_str(),voms_data);
      if(err != AAA_POSITIVE_MATCH) return err; 
    };
  };
  if(voms_data.size() == 0) return AAA_NO_MATCH;
  // analyse permissions
  for(vector<struct voms>::iterator v = voms_data.begin();v!=voms_data.end();++v) {
    // cerr<<"match vo: "<<v->voname<<endl;
    if((vo == "*") || (vo == v->voname)) {
      for(vector<struct data>::iterator d=v->std.begin();d!=v->std.end();++d) {
        // cerr<<"match group: "<<d->group<<endl;
        // cerr<<"match role: "<<d->role<<endl;
        // cerr<<"match capabilities: "<<d->cap<<endl;
        if(((group == "*") || (group == d->group)) &&
           ((role == "*") || (role == d->role)) &&
           ((capabilities == "*") || (capabilities == d->cap))) {
          default_voms_=v->server.c_str();
          default_vo_=v->voname.c_str();
          default_role_=d->role.c_str();
          default_capability_=d->cap.c_str();
          default_vgroup_=d->group.c_str();
          return AAA_POSITIVE_MATCH;
        };
      };
    };
  };
  return AAA_NO_MATCH;
}

#ifdef OLD_VOMS
int process_vomsproxy(const char* filename,vector<struct voms> &data,bool auto_cert) {
  globus_result_t res;
  globus_gsi_cred_handle_t cred;
  X509 * cert;
  STACK_OF(X509) * cert_chain;
  string voms_dir = "/etc/grid-security/vomsdir";
  string cert_dir = "/etc/grid-security/certificates";
  {
    char *v;
    if((v = getenv("X509_VOMS_DIR"))) voms_dir = v;
    if((v = getenv("X509_CERT_DIR"))) cert_dir = v;
  };
  if((res=globus_gsi_cred_handle_init(&cred,GLOBUS_NULL)) != GLOBUS_SUCCESS) {
    olog<<"Globus error: "<<res<<endl;
    return AAA_FAILURE;
  };
  if((res=globus_gsi_cred_read_proxy(cred,(char*)filename)) != GLOBUS_SUCCESS) {
    olog<<"Globus error: "<<res<<endl;
    globus_gsi_cred_handle_destroy(cred);
    return AAA_FAILURE;
  };
  if((res=globus_gsi_cred_get_cert(cred,&cert)) != GLOBUS_SUCCESS) {
    olog<<"Globus error: "<<res<<endl;
    globus_gsi_cred_handle_destroy(cred);
    return AAA_FAILURE;
  };
  if((res=globus_gsi_cred_get_cert_chain(cred,&cert_chain)) != GLOBUS_SUCCESS) {
    olog<<"Globus error: "<<res<<endl;
    globus_gsi_cred_handle_destroy(cred);
    return AAA_FAILURE;
  };
  string message;
  string vo;
  string file;
  string subject;
  string ca;
  verror_type err;
  if(!retrieve(cert,cert_chain,RECURSE_CHAIN,message,vo,file,subject,ca,err)) {
    olog<<"VOMS error: "<<err<<" - "<<voms_error(err)<<endl;
    globus_gsi_cred_handle_destroy(cred);
    if(err == VERR_NOEXT) return AAA_NO_MATCH;
    return AAA_FAILURE;
  };
  // validate information
  // first stage - parse data
  if(auto_cert) {
    struct vomsdata vdata;
    if(!verify(message,vdata,err,voms_dir,cert_dir,subject,ca,false)) {
      olog<<"VOMS parsing error: "<<err<<" - "<<voms_error(err)<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    for(int n = 0;n<vdata.data.size();n++) {
      if(vdata.data[n].uri.empty() || vdata.data[n].server.empty()) {
        olog<<"Missing URI or server subject"<<endl;
        globus_gsi_cred_handle_destroy(cred);
        return AAA_FAILURE;
      };
      if(renew_voms_cert(vdata.data[n].server,voms_dir,vdata.data[n].uri) != AAA_POSITIVE_MATCH) {
        olog<<"Certificate renewal failed"<<endl;
        globus_gsi_cred_handle_destroy(cred);
        return AAA_FAILURE;
      };
    };
  };
  // second stage - verify data
  {
    struct vomsdata vdata;
    if(!verify(message,vdata,err,voms_dir,cert_dir,subject,ca,false)) {
      olog<<"VOMS verification error: "<<err<<" - "<<voms_error(err)<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    for(int n = 0;n<vdata.data.size();n++) {
      data.push_back(vdata.data[n]);
    };
  };
/*
  do {
    struct ext_voms unformatted;
    if(!unformat(message,unformatted,err)) {
      olog<<"VOMS error: "<<err<<" - "<<voms_error(err)<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    if(unformatted.voms.uri.empty() || unformatted.voms.server.empty()) {
      olog<<"Missing URI or server subject"<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    if(renew_voms_cert(unformatted.voms.server,voms_dir,unformatted.voms.uri) != AAA_POSITIVE_MATCH) {
      olog<<"Certificate renewal failed"<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    if (!check(unformatted, err, voms_dir, cert_dir)) {
      olog<<"Failed to verify signature of VOMS data: "<<voms_error(err)<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    * Parse the date. *
    ASN1_TIME *time1 = NULL, *time2 = NULL;
    unsigned char *timebuf;
    timebuf = (unsigned char *)unformatted.voms.date1.c_str();
    time1 = d2i_ASN1_TIME(NULL, &timebuf, unformatted.voms.date1.size());
    timebuf = (unsigned char *)unformatted.voms.date2.c_str();
    time2 = d2i_ASN1_TIME(NULL, &timebuf, unformatted.voms.date2.size());
    if (!time1 || !time2) {
      olog<<"Validation times are missing in VOMS data"<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    * Verifies the date constraints *
    if((X509_cmp_current_time(time1) >=  0) || 
       (X509_cmp_current_time(time2) <= 0)) {
      olog<<"VOMS data is not valid yet/already"<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    if(!parse_rest(unformatted.datalen,unformatted.real_data,unformatted.voms,err)) {
      olog<<"VOMS error: "<<err<<" - "<<voms_error(err)<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    if(unformatted.voms.voname.empty()) {
      olog<<"VO name is missing in VOMS data"<<endl;
      globus_gsi_cred_handle_destroy(cred);
      return AAA_FAILURE;
    };
    message=unformatted.extra_data;
    data.push_back(unformatted.voms); * keep info *
  } while (!message.empty());
*/
  globus_gsi_cred_handle_destroy(cred);
  return AAA_POSITIVE_MATCH;
}
#else
int process_vomsproxy(const char* filename,vector<struct voms> &data,bool auto_cert) {


  return AAA_POSITIVE_MATCH;
}
#endif

int renew_voms_cert(const string &contact,const string &voms_dir,const string &uri) {
  // try to load certificate 
//  string contact = voms.data[0].server;
  bool voms_cert_cached = false;
  // Is that certificate already cached ? 
  BIO *in = BIO_new(BIO_s_file());
  if(in == NULL) {
    olog<<"Failed BIO_new"<<endl;
    return AAA_FAILURE;
  };
  string voms_cert = contact;
  for(int p = voms_cert.find('/');p != string::npos;p=voms_cert.find('/',p)) {
    if(p == 0) { voms_cert.erase(0,1); }
    else { voms_cert.replace(p,1,","); };
  };
  voms_cert=voms_dir+"/"+voms_cert;
  // olog<<"Cached certificate: "<<voms_cert<<endl;
  if(BIO_read_filename(in,(void*)(voms_cert.c_str())) > 0) {
    X509 *x = PEM_read_bio_X509(in, NULL, 0, NULL);
    if(x)  {
      /* Is it still valid ? */
      if(X509_get_notBefore(x) && X509_get_notAfter(x)) {
        if((X509_cmp_current_time(X509_get_notBefore(x)) < 0) && 
           (X509_cmp_current_time(X509_get_notAfter(x)) > 0)) {
          voms_cert_cached=true;
        };
      };
      X509_free(x);
    };
  };
  BIO_free(in);
  if(!voms_cert_cached) {
    /* Reload certificate */
    string host = uri;
    int port = 15000;
    int p = host.find(':'); 
    if(p != string::npos) {
      stringtoint(host.substr(p+1,string::npos),port);
      host.erase(p,string::npos);
    };
    // olog<<"Host: "<<host<<endl;
    // olog<<"Port: "<<port<<endl;
    GSISocketClient sock(host,port,22);
    sock.RedirectGSIOutput(stderr);
    sock.ServerContact(contact);
    sock.SetFlags(GSS_C_MUTUAL_FLAG | GSS_C_CONF_FLAG | GSS_C_INTEG_FLAG);
    if (!sock.Open()) {
      olog<<"Failed to open connection to "<<host<<":"<<port<<endl;
      return AAA_FAILURE;
    };
    if(sock.peer_cert == NULL) {
      olog<<"Failed to obtain peer certificate from "<<host<<":"<<port<<endl;
      return AAA_FAILURE;
    };
    BIO *out = BIO_new(BIO_s_file());
    if(out == NULL) {
      olog<<"Failed BIO_new"<<endl;
      return AAA_FAILURE;
    };
    if(BIO_write_filename(out,(void*)(voms_cert.c_str())) <= 0) {
      olog<<"Failed BIO_write_filename"<<endl;
      return AAA_FAILURE;
    };
    PEM_write_bio_X509(out,sock.peer_cert);
    BIO_free(out);
    sock.Close();
  }; 
  return AAA_POSITIVE_MATCH;
};

const char* voms_errors[] = {
  "Success",
  "Socket problem",
  "Cannot identify itself (certificate problem)",
  "Server problem",
  "Wrong parameters",
  "VOMS extension missing",
  "Initialization error",
  "Error in time checking",
  "User data in extension different from the real ones",
  "VO name and URI missing",
  "Wrong data format",
  "Empty extension",
  "Parse error",
  "Directory error",
  "Signature error",
  "Unidentifiable VOMS server",
  "Memory problems",
  "Generic verification error",
  "Returned data of unknown type"
};

const char* voms_error(verror_type err) {
  if((VERR_NONE < 0) || (err > VERR_TYPE)) return "Unknown";
  return voms_errors[err];
}




